# Parallel Python Software: http://www.parallelpython.com
# Copyright (c) 2005-2007, Vitalii Vanovschi
# All rights reserved.
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice, 
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation 
#      and/or other materials provided with the distribution.
#    * Neither the name of the author nor the names of its contributors 
#      may be used to endorse or promote products derived from this software 
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
# THE POSSIBILITY OF SUCH DAMAGE.
"""Parallel Python Software, Execution Server

http://www.parallelpython.com - updates, documentation, examples and support forums
"""

import os, thread, popen2, inspect, sys, logging, socket, sha, types, time
import cPickle as pickle
import pptransport

copyright = "Copyright (c) 2005-2007 Vitalii Vanovschi. All rights reserved"
version = "1.4.1"

class _Task:
    def __init__(self, server, tid, callback=None, callbackargs=(), group='default'):
        self.lock = thread.allocate_lock()
        self.lock.acquire()
        self.tid = tid
        self.callback = callback
        self.callbackargs = callbackargs
        self.group = group
        self.finished = False
        self.unpickled = False

    # for internal use only
    def finalize(self, sresult):
        self.sresult = sresult
        if self.callback:
            self.__unpickle()
        self.lock.release()
         

    def __call__(self, raw_result=False):
        self.wait()

        if not self.unpickled and not raw_result:
            self.__unpickle()

        if raw_result:
            return self.sresult
        else:
            return self.result
    
    def wait(self):
        if not self.finished:
            self.lock.acquire() 
            self.lock.release() 
            self.finished = True
    

    def __unpickle(self):
        self.result, sout = pickle.loads(self.sresult)
        self.unpickled = True
        if len(sout) > 0:
            print sout,
        if self.callback:
            args = self.callbackargs + (self.result,)
            self.callback(*args)

class _Worker:
    def __init__(self):
        command = "\"" + sys.executable + "\" \"" \
                + os.path.dirname(os.path.abspath(__file__)) + os.sep+"ppworker.py\""
        #workaround for Windows
        if sys.platform.startswith("win"):
            command = "\"" + command + "\""
        self.t = pptransport.CPipeTransport(*popen2.popen2(command))
        self.pid = int(self.t.receive())
        self.is_free = True

class _RWorker(pptransport.CSocketTransport):
    def __init__(self, host, port, secret):
        pptransport.SocketTransport.__init__(self)
        self.id = host + ":" + str(port)
        self.is_free = True
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((host, port))
        self.authenticate(secret)

    def __del__(self):
        logging.debug("Closing client socket")
        self.socket.close()


class _Statistics:
    def __init__(self):
        self.time = 0.0
        self.njobs = 0
        self.socket = None

class Server: 
    """Parallel Python SMP execution server class
    """

    default_port = 60000
    default_secret = "epo20pdosl;dksldkmm"

    def __init__(self, ncpus="autodetect", ppservers=(), secret=None, loglevel=logging.WARNING, 
            logstream=sys.stderr):
        """ Creates server instance
    
            ncpus - the number of worker processes, if parameter is omitted 
                    it will be set to the number of processors in the system
            ppservers - list of active parallel python execution servers to connect with
            secret - passphrase for network connections, if omitted a default passphrase will be used.
                    It's highly recommended to use a custom passphrase for all network connections.
            loglevel - logging level
            logstream - log destiunation
            
            With ncpus = 1 all tasks are executed consequently
            For the best performance either use the default "autodetect" value 
            or set ncpus to the total number of processors in the system
        """
        self.__initLog(loglevel, logstream)
        logging.debug("Creating server instance " + str(self))
        self.__tid = 0
        self.__active_tasks = 0
        self.__active_tasks_lock = thread.allocate_lock() 
        self.__queue = []   
        self.__queue_lock = thread.allocate_lock()
        self.__workers = []
        self.__rworkers = []
        self.__rworkers_reserved = []
        self.__rworkers_reserved4 = []
        self.__sourcesHM = {}
        self.__sfuncHM = {}
        self.__waittasks = []
        self.__waittasks_lock = thread.allocate_lock()
        self.set_ncpus(ncpus)
        self.ppservers = ppservers
        self.__stats = {"local":_Statistics()}
        self.__stats_lock = thread.allocate_lock()
        for ppserver in ppservers:
            if len(ppserver.split(":")) < 2:
                ppserver += ":"+str(Server.default_port)
            self.__stats[ppserver] = _Statistics()
        if secret is not None:
            if not isinstance(secret, types.StringType):
                raise TypeError("secret must be of a string type")
            self.secret = str(secret)
        else:
            self.secret = Server.default_secret
        self.__connect()
        self.__creation_time = time.time()
        logging.info("ppsmp server started with %d workers" % (self.__ncpus,))

    def get_stats(self):
        """Returns job execution statistics as a dictionary"""
        for node, stat in self.__stats.items():
            if stat.socket:
                stat.socket.send("TIME")
                stat.time = float(stat.socket.receive())
        return self.__stats

    def print_stats(self):
        """Prints job execution statistics. Useful for benchmarking on clusters"""
        print "Job execution statistics:"
        walltime = time.time()-self.__creation_time
        statistics = self.get_stats().items()
        totaljobs = 0.0
        for ppserver, stat in statistics:
            totaljobs += stat.njobs
        print " job count | % of all jobs | job time sum | time per job | job server"
        for ppserver, stat in statistics:
            if stat.njobs:
                print "    %6i |        %6.2f |     %8.4f |  %11.6f | %s" \
                        % (stat.njobs, 100.0*stat.njobs/totaljobs, stat.time, stat.time/stat.njobs, ppserver, )
        print "Time elapsed since server creation", walltime
        print

    def __connect(self):
        for ppserver in self.ppservers:
            ppserver = ppserver.split(":")
            host = ppserver[0]
            if len(ppserver)>1:
                port = int(ppserver[1])
            else:
                port = Server.default_port
            csocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                csocket.connect((host, port))
            except:
                logging.warning("Connection refused:"\
                        + " ppserver host=%s,  port=%i" % (host, port))
                continue 
            socket1 = pptransport.SocketTransport(csocket)
            #authenticate
            if not socket1.authenticate(self.secret):
                logging.warning("Authentication failed:"\
                        + " ppserver host=%s,  port=%i" % (host, port))
                continue 
            #request statistics
            socket1.send("STAT")
            ncpus = int(socket1.receive())
            self.__stats[host+":"+str(port)].socket = socket1 

            logging.debug("Connected to ppserver (host=%s, port=%i) with %i workers" 
                    % (host, port, ncpus))
            for x in range(ncpus):
                rworker = _RWorker(host, port, self.secret)
                rworker.send("EXEC")
                self.__rworkers.append(rworker)
            #creating reserved rworkers
            for x in range(ncpus):
                rworker = _RWorker(host, port, self.secret)
                rworker.send("EXEC")
                self.__rworkers_reserved.append(rworker)
            #creating reserved4 rworkers
            for x in range(ncpus*0):
                rworker = _RWorker(host, port, self.secret)
                rworker.send("EXEC")
                self.__rworkers_reserved4.append(rworker)
    
    def __del__(self):
        logging.debug("Deleting server instance " + str(self))
        for worker in self.__workers:
            logging.debug("Exiting worker process with pid=%d " % (worker.pid,))
            if sys.platform.startswith("win"):
                os.popen('TASKKILL /PID '+str(worker.pid)+' /F')
            else:
                os.kill(worker.pid, 9)
                os.waitpid(worker.pid, 0)

    def __detect_ncpus(self):
        #for Unix and Macos
        if hasattr(os, "sysconf"):
            ncpus = os.sysconf("SC_NPROCESSORS_ONLN")
            if isinstance(ncpus, int) and ncpus > 0:
                return ncpus
        #for Windows
        if os.environ.has_key("NUMBER_OF_PROCESSORS"):
            ncpus = int(os.environ["NUMBER_OF_PROCESSORS"]);
            if ncpus > 0:
                return ncpus
        #return the default value
        return 1
        
    def __initLog(self, loglevel, logstream):
        logHandler = logging.StreamHandler(logstream)
        logHandler.setLevel(loglevel)
        logHandler.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(message)s'))
        self.__logger = logging.getLogger('')
        self.__logger.addHandler(logHandler)
        self.__logger.setLevel(loglevel)

    def __dumpsfunc(self, funcs, modules):
        hashs =  hash(funcs+modules)
        if not self.__sfuncHM.has_key(hashs):
            sources = [self.__get_source(func) for func in funcs]
            self.__sfuncHM[hashs] = pickle.dumps((funcs[0].func_name, sources, modules))
        return self.__sfuncHM[hashs]

    #recursively finds all the modules in dict
    def __find_modules(self, prefix, dict):
        modules = []
        for name, object in dict.items():
            if isinstance(object, types.ModuleType) and name not in ("__builtins__", "pp"):
                if object.__name__ == prefix+name or prefix == "":
                    modules.append(object.__name__)
                    modules.extend(self.__find_modules(object.__name__+".", object.__dict__))
        return modules


    def submit(self, func, args=(), depfuncs=(), modules=(), callback=None, callbackargs=()
            , group='default',globals=None):
        """Submits function to the execution queue

            func - function to be executed
            args - tuple with arguments of the 'func'
            depfuncs - tuple with functions which might be called from 'func'
            modules - tuple with module names to import
            callback - callback function which will be called with argument list equal to callbackargs+(result,) 
                    as soon as calculation is done
            callbackargs - additional arguments for callback function
            group - job group, is used when wait(group) is called to wait for jobs in a given group to finish
            globals - dictionary from which all modules and functions will be imported, for instance: globals=globals()
            
        """
        tid = self.__gentid()
        if globals:
            modules = tuple(set(self.__find_modules("", globals)))
            self.__logger.debug("Task %i will autoimport next modules: %s" % (tid, str(modules)))
            for name, object in globals.items():
                if isinstance(object, types.FunctionType):
                     depfuncs += (object, )
        task = _Task(self, tid, callback, callbackargs, group)

        self.__waittasks_lock.acquire()
        self.__waittasks.append(task)
        self.__waittasks_lock.release()

        #if the function is a method of a class add self to the arguments list
        if isinstance(func, types.MethodType):
            args = (func.im_self,) + args

        #if there is an instance of a user deined class in the arguments add whole calss to dependancies  
        for arg in args:
            if isinstance(arg, types.InstanceType):
                depfuncs += (arg.__class__, )

        #if there is an instance of a function in the arguments add this function to dependancies
        for arg in args:
            if isinstance(arg, types.FunctionType):
                depfuncs += (arg, )

        sfunc = self.__dumpsfunc((func,)+depfuncs, modules)
        sargs = pickle.dumps(args)

        self.__queue_lock.acquire()
        self.__queue.append((task, sfunc, sargs))
        self.__queue_lock.release()

        self.__logger.debug("Task %i submited, function='%s'" % (tid, func.func_name))
        self.__scheduler()
        return task

    def wait(self, group=None):
        """Waits for all jobs in a given group to finish.
           If group is omitted waits for all jobs to finish
        """
        while True:
            self.__waittasks_lock.acquire()
            for task in self.__waittasks:
                if not group or task.group == group:
                    self.__waittasks_lock.release()
                    task.wait()
                    break
            else:
                self.__waittasks_lock.release()
                break

    def insert(self, sfunc, sargs):
        """Inserts function into the execution queue. It's intended for internal use only (ppserver.py)."""
        tid = self.__gentid()
        task = _Task(self, tid)
        self.__queue_lock.acquire()
        self.__queue.append((task, sfunc, sargs))
        self.__queue_lock.release()
        self.__logger.debug("Task %i inserted" % (tid, ))
        self.__scheduler()
        return task        

    def get_ncpus(self):
        """Returns the number of worker processes"""
        return self.__ncpus

    def set_ncpus(self, ncpus="autodetect"):
        """Sets the number of worker processes
        
        ncpus - the number of worker processes, if parammeter is omitted
                it will be set to the number of processors in the system"""
        if ncpus == "autodetect":
            ncpus = self.__detect_ncpus()
        if not isinstance(ncpus, int):
            raise TypeError("ncpus must have 'int' type")
        if ncpus < 0:
            raise ValueError("ncpus must be an integer > 0");
        if ncpus > len(self.__workers):
            self.__workers.extend([_Worker() for x in range(ncpus - len(self.__workers))])
        self.__ncpus = ncpus

    def __scheduler(self):
        self.__queue_lock.acquire()
        while self.__queue:
            if self.__active_tasks < self.__ncpus:
                #TODO: select a job number on the basis of heuristic instead of 0
                task = self.__queue.pop(0)
                for worker in self.__workers:
                    if worker.is_free:
                        worker.is_free = False
                        break
                else:
                    self.__logger.error("There are no free workers left")
                    raise "Error: No free workers"
                self.__add_to_active_tasks(1)
                try:
                    self.__stats["local"].njobs += 1
                    thread.start_new_thread(self.__run, task+(worker,))
                except:
                    pass
            else: 
                for rworker in self.__rworkers:
                    if rworker.is_free:
                        rworker.is_free = False
                        task = self.__queue.pop(0)
                        self.__stats[rworker.id].njobs += 1
                        thread.start_new_thread(self.__rrun, task+(rworker,))
                        break
                else:
                    if len(self.__queue) > self.__ncpus:
                        for rworker in self.__rworkers_reserved:
                            if rworker.is_free:
                                rworker.is_free = False
                                task = self.__queue.pop(0)
                                self.__stats[rworker.id].njobs += 1
                                thread.start_new_thread(self.__rrun, task+(rworker,))
                                break
                        else:
                            break
                            # this code will not be executed
                            # and is left for further releases
                            if len(self.__queue) > self.__ncpus*0:
                                for rworker in self.__rworkers_reserved4:
                                    if rworker.is_free:
                                        rworker.is_free = False
                                        task = self.__queue.pop(0)
                                        self.__stats[rworker.id].njobs += 1
                                        thread.start_new_thread(self.__rrun, task+(rworker,))
                                        break
                    else:
                        break
    
        self.__queue_lock.release()
         
    def __get_source(self, func):
        hashf = hash(func)
        if not self.__sourcesHM.has_key(hashf):
            #get lines of the source and adjust indent
            sourcelines = inspect.getsourcelines(func)[0]
            fl =  sourcelines[0]
            i = 0
            while fl[i] in (' ', '\t'):
                i += 1
            source = ""
            for line in sourcelines:
                source += line[i:]
#            source = inspect.getsource(func)
            
            self.__sourcesHM[hashf] = source
        return self.__sourcesHM[hashf]

    def __rrun(self, job, sfunc, sargs, rworker):
        self.__logger.debug("Task (remote) %i started" % (job.tid,))

        rworker.csend(sfunc)
        rworker.send(sargs)
        sresult = rworker.receive()
        rworker.is_free = True

        job.finalize(sresult)

        # remove the job from the waiting list
        if self.__waittasks:
            self.__waittasks_lock.acquire()
            self.__waittasks.remove(job)
            self.__waittasks_lock.release()

        self.__logger.debug("Task (remote) %i ended" % (job.tid,))
        self.__scheduler()

    def __run(self, job, sfunc, sargs, worker):
        self.__logger.debug("Task %i started" % (job.tid,))

        start_time = time.time()

        worker.t.csend(sfunc)
        worker.t.send(sargs)
        sresult = worker.t.receive()
        worker.is_free = True
        
        job.finalize(sresult)

        # remove the job from the waiting list
        if self.__waittasks:
            self.__waittasks_lock.acquire()
            self.__waittasks.remove(job)
            self.__waittasks_lock.release()

        self.__add_to_active_tasks(-1)

        self.__stat_add_time("local", time.time()-start_time)

        self.__logger.debug("Task %i ended" % (job.tid,))
        self.__scheduler()

    def __add_to_active_tasks(self, num):
        self.__active_tasks_lock.acquire()
        self.__active_tasks += num
        self.__active_tasks_lock.release()

    def __stat_add_time(self, node, time):
        self.__stats_lock.acquire()
        self.__stats[node].time += time
        self.__stats_lock.release()

    def __stat_add_job(self, node):
        self.__stats_lock.acquire()
        self.__stats[node].njobs += 1
        self.__stats_lock.release()

    def __gentid(self):
        self.__tid += 1
        return self.__tid - 1

# Parallel Python Software: http://www.parallelpython.com
